pub use std::error::Error;

#[derive(Debug)]
pub enum ServiceError {
    NoPermission,
    NotImplementedYet,
    AlreadyStarted,
    AlreadyStopped,
    Failed,
    IOError(std::io::Error)
}

impl Error for ServiceError {
    fn source(&self) -> Option<&(dyn Error + 'static)> {
        match self {
            Self::IOError(err) => Some(err),
            _ => None
        }
    }
}

impl From<std::io::Error> for ServiceError {
    fn from(err: std::io::Error) -> Self {
        ServiceError::IOError(err)
    }
}

impl core::fmt::Display for ServiceError {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "Service Error: ")
            .and(match self {
                Self::NoPermission => write!(f, "no permission"),
                Self::NotImplementedYet => write!(f, "not implemented yet"),
                Self::AlreadyStarted => write!(f, "already started"),
                Self::AlreadyStopped => write!(f, "already stopped"),
                Self::Failed => write!(f, "failed"),
                Self::IOError(err) => write!(f, "IO error: {}", err)
            })
    }
}

#[derive(Debug, Copy, Clone)]
pub enum Service {
    Daemon,
    Minecraft
}

impl Service {
    pub async fn start(&self) -> Result<(), ServiceError> {
        match self {
            Self::Daemon => Err(ServiceError::AlreadyStarted),
            #[cfg(not(feature = "minecraft-server"))]
            Self::Minecraft => Err(ServiceError::NotImplementedYet),
            #[cfg(feature = "minecraft-server")]
            Self::Minecraft => if !self.is_running().await {
                // Configure the command
                let res = tokio::process::Command::new("screen")
                    .current_dir("/media/games/bukkit")
                    .args(r#"-dmS sfs-mc -t sfs-mc "java -server -Xmx2G -jar server.jar""#.split_whitespace())
                    // Handle spawning
                    .spawn()
                    .map_err(|err| ServiceError::IOError(err))?
                    .await
                    // Convert io::Error to ServiceError
                    .map_err(|err| err.into());

                if let Ok(status) = res {
                    if !status.success() {
                        Err(ServiceError::Failed)?;
                    }
                }

                res.map(|_| ())
            } else {
                Err(ServiceError::AlreadyStarted)
            }
        }
    }

    pub async fn is_running(&self) -> bool {
        match self {
            Self::Daemon => true,
            #[cfg(not(feature = "minecraft-server"))]
            Self::Minecraft => false,
            #[cfg(feature = "minecraft-server")]
            Self::Minecraft => tokio::process::Command::new("screen")
                .args("-rqx sfs-mc -t sfs-mc -Q title".split_whitespace())
                .output().await
                .unwrap()
                .status
                .success()
        }
    }

    pub async fn stop(&self) -> Result<(), ServiceError> {
        match self {
            Self::Daemon => std::process::exit(0),
            #[cfg(not(feature = "minecraft-server"))]
            Self::Minecraft => Err(ServiceError::NotImplementedYet),
            #[cfg(feature = "minecraft-server")]
            Self::Minecraft => Err(ServiceError::NoPermission)
        }
    }

    pub async fn restart(&self) -> Result<(), ServiceError> {
        match self {
            Self::Daemon => Err(ServiceError::NoPermission),
            _ => {
                if self.is_running().await {
                    self.stop().await?;
                }
                self.start().await
            }
        }
    }
}

#[derive(Default, Debug, Copy, Clone)]
pub struct ServiceData;

impl dbus_tree::DataType for ServiceData {
    type Tree = ();
    type ObjectPath = std::sync::Arc<tokio::sync::RwLock<Service>>;
    type Property = ();
    type Interface = ();
    type Method = ();
    type Signal = ();
}

#[macro_export]
macro_rules! handle_iface {
    ($m:ident, |$data:ident| $($f:tt)+) => {move |$m| tokio::task::block_in_place(move || {
        let $data = $m.path.get_data();
        // Run a Future on the dedicated tokio runtime
        iface_rt().block_on(async move {
            $($f)+
        })
    })};
}